@@ -355,11 +355,21 @@ func FromFPointsToSamples(points []promql.FPoint) []Sample {
 	return *(*[]Sample)(unsafe.Pointer(&points))
 }
 
+// FromSamplesToFPoints casts []Sample to []promql.FPoint. It uses unsafe.
+func FromSamplesToFPoints(samples []Sample) []promql.FPoint {
+	return *(*[]promql.FPoint)(unsafe.Pointer(&samples))
+}
+
 // FromHPointsToHistograms converts []promql.HPoint to []FloatHistogramPair. It uses unsafe.
 func FromHPointsToHistograms(points []promql.HPoint) []FloatHistogramPair {
 	return *(*[]FloatHistogramPair)(unsafe.Pointer(&points))
 }
 
+// FromHistogramsToHPoints converts []FloatHistogramPair to []promql.HPoint. It uses unsafe.
+func FromHistogramsToHPoints(histograms []FloatHistogramPair) []promql.HPoint {
+	return *(*[]promql.HPoint)(unsafe.Pointer(&histograms))
+}
+
 // FromFloatHistogramToPromHistogram converts histogram.FloatHistogram to model.SampleHistogram.
 func FromFloatHistogramToPromHistogram(h *histogram.FloatHistogram) *model.SampleHistogram {
 	if h == nil {
